@@ -2,10 +2,7 @@
 
 package main
 
-import (
-	"github.com/Sirupsen/logrus"
-	"github.com/codegangsta/cli"
-)
+import "github.com/codegangsta/cli"
 
 var (
 	checkpointCommand cli.Command
@@ -14,7 +11,3 @@ var (
 	specCommand       cli.Command
 	killCommand       cli.Command
 )
-
-func runAction(*cli.Context) {
-	logrus.Fatal("Current OS is not supported yet")
-}
